from pydantic import BaseModel

from .quantity import Quantity

try:
    from .pandas_engine import QuantityDtype
except ModuleNotFoundError:
    QuantityDtype = None

BaseModel.__getitem__ = lambda self, key: self.__dict__.get(key)

__all__ = ["Quantity", "QuantityDtype"]



from __future__ import annotations

import re
from typing import ClassVar, Literal

import numpy as np
import numpy.typing as npt
import pandas as pd
from pandas.core import ops
from pandas.core.algorithms import take
from pandas.core.arrays.numeric import NumericArray, NumericDtype
from pandas.core.dtypes.base import register_extension_dtype
from pandas.core.dtypes.common import is_bool, is_float_dtype, pandas_dtype
from pandas.core.dtypes.missing import isna, notna
from pandas.core.indexers import check_array_indexer
from pandas.util._decorators import cache_readonly

from .quantity import ISUnits, Quantity


@register_extension_dtype
class QuantityDtype(NumericDtype):
    """Pandas dtype for quantities."""

    type = np.float64
    name: ClassVar[str] = "Quantity"
    _metadata = ("unit",)
    _default_np_dtype = np.dtype(np.float64)
    _checker = is_float_dtype

    @property
    def name(self) -> str:  # noqa: F811
        """Dtype display name."""
        return f"Quantity[{self.unit}]"

    def __init__(self, unit: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.unit = unit

    def __repr__(self):
        """Dtype repr."""
        return f"QuantityDtype({self.unit})"

    @classmethod
    def construct_array_type(cls) -> type[QuantityArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        return QuantityArray

    @classmethod
    def construct_from_string(cls, string: str):
        """Construct QuantityDtype from string, used when retrieving from arrow."""
        try:
            unit = re.findall(r"^Quantity\[(.*)\]$", string)[0]
            return cls(unit=unit)
        except Exception as exc:
            raise TypeError(f"Cannot construct QuantityDtype from {string}") from exc

    @classmethod
    def _safe_cast(cls, values: np.ndarray, dtype: np.dtype, copy: bool) -> np.ndarray:
        """
        Safely cast the values to the given dtype.

        "safe" in this context means the casting is lossless.
        """
        # This is really only here for compatibility with IntegerDtype
        # Here for compat with IntegerDtype
        return values.astype(dtype, copy=copy)

    def __from_arrow__(self, array, **kwargs):
        """Construct QuantityArray from pyarrow Array."""
        import pyarrow
        from pandas.core.arrays.arrow._arrow_utils import pyarrow_array_to_numpy_and_mask

        array_class = self.construct_array_type()

        pyarrow_type = pyarrow.from_numpy_dtype(self.type)
        if not array.type.equals(pyarrow_type) and not pyarrow.types.is_null(array.type):
            # test_from_arrow_type_error raise for string, but allow
            #  through itemsize conversion GH#31896
            rt_dtype = pandas_dtype(array.type.to_pandas_dtype())
            if rt_dtype.kind not in "iuf":
                # Could allow "c" or potentially disallow float<->int conversion,
                #  but at the moment we specifically test that uint<->int works
                raise TypeError(f"Expected array of {self} type, got {array.type} instead")

            array = array.cast(pyarrow_type)

        if isinstance(array, pyarrow.ChunkedArray):
            # TODO this "if" can be removed when requiring pyarrow >= 10.0, which fixed
            # combine_chunks for empty arrays https://github.com/apache/arrow/pull/13757
            array = pyarrow.array([], type=array.type) if array.num_chunks == 0 else array.combine_chunks()

        data, mask = pyarrow_array_to_numpy_and_mask(array, dtype=self.numpy_dtype)
        return array_class(self.unit, data.copy(), ~mask, copy=False)

    def _get_common_dtype(self, dtypes: list):
        if not all(isinstance(dtype, QuantityDtype) for dtype in dtypes):
            raise ValueError("No common dtype between Quantity and non-Quantity types")
        raise ValueError("No common dtype between Quantity with different units")

    def __eq__(self, other):
        """Equality between QuantityDtypes."""
        return isinstance(other, QuantityDtype) and self.unit == other.unit

    def __hash__(self):
        """Hash function."""
        return hash(str(self))


class QuantityArray(NumericArray):
    """Pandas array for quantities."""

    _dtype_cls = QuantityDtype
    _internal_fill_value = np.nan

    @cache_readonly
    def dtype(self) -> QuantityDtype:
        """Pandas dtype associated with this value."""
        return QuantityDtype(unit=self.unit)

    def __init__(self, unit: str, values: np.ndarray, mask: npt.NDArray[np.bool_], copy: bool = False):
        self.unit = unit
        super().__init__(values, mask, copy=copy)

    @classmethod
    def _from_sequence(cls, scalars, *, dtype: QuantityDtype, copy: bool = False):
        values, mask = cls._coerce_to_array(scalars, dtype=dtype, copy=copy)
        return cls(unit=dtype.unit, values=values, mask=mask)

    def _arith_method(self, other, op):
        pd_op = ops.get_array_op(op)
        self_q = Quantity(value=self._data, unit=self.dtype.unit)
        mask = self._mask
        if isinstance(other, QuantityArray):
            other_q = Quantity(value=other._data, unit=other.dtype.unit)
            mask |= other._mask
        elif isinstance(other, Quantity | float | int):
            other_q = other
        elif isinstance(other, np.ndarray):
            other_q = (
                Quantity(value=[x.value for x in other], unit=other[0].unit)
                if len(other) and isinstance(other[0], Quantity)
                else Quantity(value=other, unit="")
            )
        else:
            raise NotImplementedError(f"Unsupported type: {type(other)}")
        result = pd_op(self_q, other_q)
        if not isinstance(result, Quantity):
            result = Quantity(value=result, unit="")
        return self._maybe_mask_result(result, mask)

    def _maybe_mask_result(self, result: Quantity, mask: np.ndarray):
        if not isinstance(result, Quantity):
            result = Quantity(value=result, unit=self.unit)
        if not isinstance(result.value, np.ndarray):
            raise NotImplementedError("Unsupported type: {type(result.value)}")
        return QuantityArray(result.unit, result.value, mask, copy=False)

    def copy(self):
        """Return a copy of the array."""
        data = self._data.copy()
        mask = self._mask.copy()
        return self.__class__(unit=self.unit, values=data, mask=mask)

    def to(self, other_unit: str):
        """Convert to another unit."""
        new_value = Quantity(value=self._data, unit=self.unit).to(other_unit).value
        if not isinstance(new_value, np.ndarray):
            raise NotImplementedError("Unsupported type: {type(new_value)}")
        return self.__class__(unit=other_unit, values=new_value, mask=self._mask)

    def __repr__(self):
        """QuantityArray repr."""
        return repr(Quantity(value=self._data, unit=self.unit))

    def __getitem__(self, item):
        """Get item from QuantityArray."""
        item = check_array_indexer(self, item)

        newmask = self._mask[item]
        if is_bool(newmask):
            # This is a scalar indexing
            if newmask:
                return self.dtype.na_value
            return Quantity(value=self._data[item], unit=self.unit)

        return self.__class__(self.unit, self._data[item], newmask)

    def take(
        self,
        indexer,
        *,
        allow_fill: bool = False,
        fill_value=None,
        axis: Literal[0, 1] = 0,
    ):
        """Take elements from an array along an axis."""
        # we always fill with 1 internally
        # to avoid upcasting
        data_fill_value = self._internal_fill_value if isna(fill_value) else fill_value
        result = take(
            self._data,
            indexer,
            fill_value=data_fill_value,
            allow_fill=allow_fill,
            axis=axis,
        )

        mask = take(self._mask, indexer, fill_value=True, allow_fill=allow_fill, axis=axis)

        # if we are filling
        # we only fill where the indexer is null
        # not existing missing values
        # TODO(jreback) what if we have a non-na float as a fill value?
        if allow_fill and notna(fill_value):
            fill_mask = np.asarray(indexer) == -1
            result[fill_mask] = fill_value
            mask = mask ^ fill_mask

        return self.__class__(self.unit, result, mask)

    @classmethod
    def _concat_same_type(
        cls,
        to_concat,
        axis=0,
    ):
        unit = to_concat[0].unit
        if not all(x.unit == unit for x in to_concat):
            raise ValueError("Can only concatenate Quantity with the same unit")
        data = np.concatenate([x._data for x in to_concat], axis=axis)
        mask = np.concatenate([x._mask for x in to_concat], axis=axis)
        return cls(unit, data, mask)


@pd.api.extensions.register_series_accessor("qt")
class QuantitySeriesAccessor:
    """Quantity accessor for Series."""

    def __init__(self, pandas_obj: pd.Series):
        self._obj = pandas_obj

    def to(self, other_unit: str):
        """Convert to another (compatible) unit."""
        if not isinstance(self._obj.dtype, QuantityDtype):
            raise TypeError("Can only convert a Series of Quantity")
        new_value = Quantity(value=self._obj.to_numpy(), unit=self._obj.dtype.unit).to(other_unit).value
        return pd.Series(new_value, index=self._obj.index, dtype=QuantityDtype(unit=other_unit))


@pd.api.extensions.register_dataframe_accessor("qt")
class QuantityDataFrameAccessor:
    """Quantity accessor for DataFrame."""

    def __init__(self, pandas_obj: pd.DataFrame):
        self._obj = pandas_obj

    def to(self, other_unit: str):
        """Convert to another (compatible) unit."""
        if not all(isinstance(dtype, QuantityDtype) for dtype in self._obj.dtypes.values):
            raise ValueError("Can only convert a DataFrame of Quantity")
        if len(set(self._obj.dtypes.values)) != 1:
            raise ValueError("Can only convert a DataFrame of Quantity with the same unit")

        unit = self._obj.dtypes.values[0].unit
        new_value = Quantity(value=self._obj.to_numpy(), unit=unit).to(other_unit).value
        return pd.DataFrame(
            new_value, index=self._obj.index, columns=self._obj.columns, dtype=QuantityDtype(unit=other_unit)
        )

    def find(self, unit: str | None = None, category: str | None = None, i_s_units: ISUnits | str | None = None):
        """Find Quantity columns."""
        if not any([unit, category, i_s_units]):
            return self._obj[[col for col, dtype in self._obj.dtypes.items() if isinstance(dtype, QuantityDtype)]]

        if unit:
            return self._obj[self._obj.dtypes[self._obj.dtypes == QuantityDtype(unit)].index]
        if isinstance(i_s_units, str):
            i_s_units = ISUnits.from_str(i_s_units)
        cols = []
        for col, dtype in self._obj.dtypes.items():
            if isinstance(dtype, QuantityDtype):
                col_i_s_units, _, col_category = Quantity.get_unit_info(dtype.unit)
                if (category and col_category == category) or (i_s_units and col_i_s_units == i_s_units):
                    cols.append(col)
        return self._obj[cols]



from .common import (
    deep_diff,
    deep_merge,
    get_all_subclasses,
    get_model_cls,
    get_non_null_annotation,
    is_subclass,
)
from .path import (
    SEP,
    from_form_data,
    get_fullpath,
    get_model_value,
    get_subitem,
    get_subitem_cls,
    handle_discriminated,
    model_construct_recursive,
    set_at_path,
)
from .quantity import Quantity, QuantityDtype
from .types import Type

__all__ = [
    "deep_diff",
    "deep_merge",
    "from_form_data",
    "get_all_subclasses",
    "get_fullpath",
    "get_model_cls",
    "get_model_value",
    "get_non_null_annotation",
    "get_subitem",
    "get_subitem_cls",
    "handle_discriminated",
    "is_subclass",
    "model_construct_recursive",
    "Quantity",
    "QuantityDtype",
    "SEP",
    "set_at_path",
    "Type",
]



from __future__ import annotations

import re
from copy import deepcopy
from functools import lru_cache
from itertools import permutations
from types import SimpleNamespace
from typing import TYPE_CHECKING, ClassVar, TypedDict, TypeVar, overload

from pydantic import BaseModel, ConfigDict, field_validator
from typing_extensions import Self

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt
    import pandas as pd

    float_array = npt.NDArray[np.floating]
else:
    try:
        import numpy as np
        import numpy.typing as npt
        import pandas as pd

        float_array = npt.NDArray[np.floating]
    except ModuleNotFoundError:

        def raise_not_found():
            """Raise ModuleNotFoundError."""
            raise ModuleNotFoundError("numpy not found")

        float_array = float
        np = SimpleNamespace(ndarray=float, array=raise_not_found)
        pd = SimpleNamespace(Series=None, DataFrame=None)

T = TypeVar("T")


PREFIX_MULTIPLIERS = {
    "n": 1e-9,
    "u": 1e-6,
    "m": 1e-3,
    "c": 1e-2,
    "d": 1e-1,
    "D": 1e1,
    "h": 1e2,
    "k": 1e3,
    "M": 1e6,
    "G": 1e9,
    "T": 1e12,
}


class ISUnits(BaseModel):
    """International system units + money unit."""

    kg: int = 0
    m: int = 0
    s: int = 0
    A: int = 0
    K: int = 0
    mol: int = 0
    cd: int = 0
    """Extend IS units to money quantities."""
    USD: int = 0

    model_config = ConfigDict(extra="forbid")

    def is_empty(self) -> bool:
        """Check if the model is empty."""
        return all(getattr(self, field) == 0 for field in self.model_fields)

    def __hash__(self):
        """The hash function."""
        return hash(frozenset(self.model_dump().items()))

    def __mul__(self, other: ISUnits) -> ISUnits:
        """Multiply two IS units."""
        return self.__class__(**{field: getattr(self, field) + getattr(other, field) for field in self.model_fields})

    def __truediv__(self, other: ISUnits) -> ISUnits:
        """Divide two IS units."""
        return self.__class__(**{field: getattr(self, field) - getattr(other, field) for field in self.model_fields})

    def __pow__(self, other: int) -> ISUnits:
        """Raise to power."""
        return self.__class__(**{field: getattr(self, field) * other for field in self.model_fields})

    def __repr__(self) -> str:
        """Representation."""
        return f"ISUnits({self})"

    def __str__(self) -> str:
        """String representation."""
        return "*".join([unit + (f"^{pow}" if pow != 1 else "") for unit, pow in self.model_dump().items() if pow])

    @classmethod
    def from_str(cls, unit_str: str) -> ISUnits:
        """Create from string."""
        return cls(
            **{
                unit: int(pow) if pow else 1
                for unit, pow in re.findall(r"(kg|m|s|K|A|mol|cd|USD)(?:\^(\-*\d+))*(?:\*|$)", unit_str)
            }
        )


class NameData(TypedDict):
    """Name data."""

    unit: str
    category: str


class Quantity(BaseModel):
    """Quantity model."""

    value: float | float_array
    """Value or array of values quantified by the unit."""
    unit: str
    """Unit of the quantity."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    __pandas_priority__ = 5000

    conversion: ClassVar[dict[ISUnits, dict[str, int | float | tuple[int | float, int | float]]]] = {
        # Unitless
        ISUnits(): {
            "": 1,
            "%": 0.01,
            "‰": 0.001,
        },
        # Money
        ISUnits(USD=1): {
            "USD": 1,
            "EUR": 1.07,
            "GBP": 1.30,
            "AUD": 0.67,
            "JPY": 0.0067,
            "CAD": 0.73,
            "CHF": 1.16,
            "CNY": 0.14,
            "NZD": 0.61,
            "ZAR": 0.057,
            "INR": 0.012,
        },
        # Length
        ISUnits(m=1): {
            "m": 1,
            "mi": 1609.344,
            "NM": 1852,
            "ft": 0.3048,
            "in": 0.0254,
        },
        # Mass
        ISUnits(kg=1): {
            "kg": 1,
            "g": 0.001,
            "lb": 0.453592,
            "oz": 0.0283495,
        },
        # Time
        ISUnits(s=1): {
            "s": 1,
            "min": 60,
            "h": 3600,
            "d": 86400,
        },
        # Force
        ISUnits(kg=1, m=1, s=-2): {
            "N": 1,
        },
        # Pressure
        ISUnits(kg=1, m=-1, s=-2): {
            "Pa": 1,
            "atm": 101325,
            "psi": 6894.76,
        },
        # Energy
        ISUnits(kg=1, m=2, s=-2): {
            "J": 1,
            "Wh": 3600,
            "cal": 4.184,
        },
        # Frequency
        ISUnits(s=-1): {
            "Hz": 1,
        },
        # Temperature
        ISUnits(K=1): {
            "K": 1,
            "C": (1, 273.15),
            "F": (5 / 9, 273.15 - 5 / 9 * 32),
        },
        # Power
        ISUnits(kg=1, m=2, s=-3): {
            "W": 1,
            "hp": 745.7,
        },
        # Area
        ISUnits(m=2): {
            "m^2": 1,
            "ha": 10000,
            "acre": 4046.86,
            "in^2": 0.00064516,
            "ft^2": 0.092903,
            "mi^2": 2589988.1,
        },
        # Volume
        ISUnits(m=3): {
            "m^3": 1,
            "L": 0.001,
            "gal": 0.00378541,
        },
        # Speed
        ISUnits(m=1, s=-1): {
            "m/s": 1,
            "mph": 0.44704,
            "kph": 1 / 3.6,
        },
        # Acceleration
        ISUnits(m=1, s=-2): {
            "m/s^2": 1,
        },
    }
    """Conversion factors from unit to IS unit."""

    names: ClassVar[dict[ISUnits, NameData]] = {
        ISUnits(): {"category": "Unitless", "unit": ""},
        ISUnits(USD=1): {"category": "Money", "unit": "USD"},
        ISUnits(m=1): {"category": "Length", "unit": "m"},
        ISUnits(kg=1): {"category": "Mass", "unit": "kg"},
        ISUnits(s=1): {"category": "Time", "unit": "s"},
        ISUnits(K=1): {"category": "Temperature", "unit": "K"},
        ISUnits(s=-1): {"category": "Frequency", "unit": "Hz"},
        ISUnits(kg=1, m=1, s=-2): {"category": "Force", "unit": "N"},
        ISUnits(kg=1, m=-1, s=-2): {"category": "Pressure", "unit": "Pa"},
        ISUnits(kg=1, m=2, s=-2): {"category": "Energy", "unit": "J"},
        ISUnits(kg=1, m=2, s=-3): {"category": "Power", "unit": "W"},
        ISUnits(m=2): {"category": "Area", "unit": "m^2"},
        ISUnits(m=3): {"category": "Volume", "unit": "m^3"},
        ISUnits(m=1, s=-1): {"category": "Speed", "unit": "m/s"},
        ISUnits(m=1, s=-2): {"category": "Acceleration", "unit": "m/s^2"},
    }
    """Category names and default units for each IS unit."""

    def __init__(self, value: float | float_array, unit: str):
        if isinstance(value, list):
            value = np.array(value)
        super().__init__(value=value, unit=unit)

    @property
    def unit_multiplier(self) -> tuple[float, float]:
        """Get multiplier."""
        return self.get_unit_info(self.unit)[1]

    @property
    def i_s_units(self) -> ISUnits:
        """Get IS units."""
        return self.get_unit_info(self.unit)[0]

    @property
    def category(self) -> str:
        """Get category."""
        return self.get_unit_info(self.unit)[2]

    @classmethod
    @lru_cache
    def get_unit_info(cls, unit: str) -> tuple[ISUnits, tuple[float, float], str]:
        """Get information about a given unit."""
        i_s_pattern = re.compile("^((kg|m|s|K|USD|A|mol|cd)(?:\^(\-*\d+))*(?:\*|$))+$")
        # IS units repr
        if re.match(i_s_pattern, unit):
            i_s_units = ISUnits.from_str(unit)
            return i_s_units, (1, 0), cls.names.get(i_s_units, {}).get("category", "Quantity")

        all_units = sum([list(units) for units in cls.conversion.values()], [])

        def trim_prefixes(unit: str) -> tuple[str, str | None]:
            if unit in all_units:
                return unit, None
            for prefix in PREFIX_MULTIPLIERS:
                if unit.startswith(prefix):
                    base_unit = unit.removeprefix(prefix)
                    if base_unit in all_units:
                        return base_unit, prefix
            raise ValueError(f"Unsupported unit: {unit}")

        i_s_units = ISUnits()
        factor, base = 1.0, 0.0
        for i, unit_ in enumerate(unit.split("/")):
            base_unit, prefix = trim_prefixes(unit_)
            i_s_units_, conversion_ = next(
                (i_s_units, group[base_unit]) for i_s_units, group in cls.conversion.items() if base_unit in group
            )
            if isinstance(conversion_, tuple):
                factor_, base_ = conversion_
                if i > 0 and base_ != 0:
                    raise NotImplementedError("Not supported: unit with non-zero base in denominator")
            else:
                factor_, base_ = conversion_, 0

            if prefix:
                prefix_multiplier = PREFIX_MULTIPLIERS[prefix]
                # Handle cases where prefix multiplier is brought to a power (e.g area, volume)
                if (
                    re.match(i_s_pattern, base_unit)
                    and sum(v != 0 for v in i_s_units_.model_dump().values()) == 1
                    and abs(pow := next(v for v in i_s_units_.model_dump().values() if v != 0)) > 1
                ):
                    prefix_multiplier = prefix_multiplier**pow

                factor_ *= prefix_multiplier

            if i == 0:
                i_s_units = i_s_units_
                factor = factor_
                base = base_
            else:
                i_s_units /= i_s_units_
                factor /= factor_

        return i_s_units, (factor, base), cls.names.get(i_s_units, {}).get("category", "Quantity")

    @property
    def unit_parts(self) -> dict[ISUnits, str]:
        """Get unit parts."""
        return {self.get_unit_info(part)[0]: part for part in self.unit.split("/")}

    @field_validator("unit")
    def validate_unit(cls, unit: str) -> str:
        """Validate unit."""
        _ = cls.get_unit_info(unit)
        return unit

    def _to_default_unit(self) -> float | float_array:
        """Convert to default unit."""
        factor, base = self.unit_multiplier
        return self.value * factor + base

    def _from_default_unit(self, value) -> float:
        """Convert from default unit."""
        factor, base = self.unit_multiplier
        return (value - base) / factor

    def to(self, unit: str) -> Self:
        """Convert to another unit."""
        if self.get_unit_info(unit)[0] != self.i_s_units:
            self_repr = self.category if self.category != "Quantity" else str(self.unit)
            other_info = self.get_unit_info(unit)
            other_repr = other_info[2] if other_info[2] != "Quantity" else str(unit)
            raise ValueError(f"Cannot convert between different types: {self_repr} <-/-> {other_repr}.")
        default_unit_value = self._to_default_unit()
        factor, base = self.get_unit_info(unit)[1]
        return self.__class__(unit=unit, value=(default_unit_value - base) / factor)

    @overload
    def __add__(self, other: pd.DataFrame) -> pd.DataFrame: ...
    @overload
    def __add__(self, other: pd.Series) -> pd.Series: ...
    @overload
    def __add__(self, other: Self | float | int | float_array) -> Self: ...
    def __add__(
        self, other: Self | float | int | float_array | pd.Series | pd.DataFrame
    ) -> Self | pd.Series | pd.DataFrame:
        """Add two quantities, they need to be of the same category."""
        if pd.DataFrame and isinstance(other, pd.DataFrame):
            return other.assign(**{col: self + other[col] for col in other.columns})

        if pd.Series and isinstance(other, pd.Series):
            from .pandas_engine import QuantityDtype

            other_unit = getattr(other.dtype, "unit", "")
            val = self + Quantity(other.to_numpy(), other_unit)
            return pd.Series(val.value, index=other.index, dtype=QuantityDtype(unit=val.unit))

        if isinstance(other, float | int | np.ndarray):
            return self.__class__(self.value + other, self.unit)

        if isinstance(other, Quantity):
            if other.i_s_units != self.i_s_units:
                raise ValueError("Cannot add quantities of different types.")

            return self.__class__(
                self._from_default_unit(self._to_default_unit() + other._to_default_unit()), self.unit
            )

        raise NotImplementedError

    @overload
    def __sub__(self, other: pd.DataFrame) -> pd.DataFrame: ...
    @overload
    def __sub__(self, other: pd.Series) -> pd.Series: ...
    @overload
    def __sub__(self, other: Self | float | int | float_array) -> Self: ...
    def __sub__(
        self, other: Self | float | int | float_array | pd.Series | pd.DataFrame
    ) -> Self | pd.Series | pd.DataFrame:
        """Add two quantities, they need to be of the same category."""
        if pd.DataFrame and isinstance(other, pd.DataFrame):
            return other.assign(**{col: self - other[col] for col in other.columns})

        if pd.Series and isinstance(other, pd.Series):
            from .pandas_engine import QuantityDtype

            other_unit = getattr(other.dtype, "unit", "")
            val = self - Quantity(other.to_numpy(), other_unit)
            return pd.Series(val.value, index=other.index, dtype=QuantityDtype(unit=val.unit))

        if isinstance(other, float | int | np.ndarray):
            return self.__class__(unit=self.unit, value=self.value + other)

        if isinstance(other, Quantity):
            if self.i_s_units != other.i_s_units:
                raise ValueError("Cannot subtract quantities of different types.")

            return self.__class__(
                unit=self.unit,
                value=self._from_default_unit(self._to_default_unit() - other._to_default_unit()),
            )

        raise NotImplementedError

    def _get_output_unit(self, other: Self | float | int | float_array, out_i_s_units: ISUnits) -> str | None:
        all_parts = deepcopy(self.unit_parts)
        if isinstance(other, Quantity):
            all_parts = deepcopy(other.unit_parts) | all_parts

        if out_i_s_units in all_parts:
            return all_parts[out_i_s_units]

        unit_parts = next(
            (
                (all_parts.get(i1) or v1["unit"], all_parts.get(i2) or v2["unit"])
                for (i1, v1), (i2, v2) in permutations(self.names.items(), 2)
                if i1 / i2 == out_i_s_units and (i1 in all_parts or i2 in all_parts)
            ),
            None,
        )

        if unit_parts:
            return "/".join(unit_parts)

        return None

    @overload
    def __mul__(self, other: pd.DataFrame) -> pd.DataFrame: ...
    @overload
    def __mul__(self, other: pd.Series) -> pd.Series: ...
    @overload
    def __mul__(self, other: Self | float | int | float_array) -> Self: ...
    def __mul__(
        self, other: Self | float | int | float_array | pd.Series | pd.DataFrame
    ) -> Self | pd.Series | pd.DataFrame:
        """Multiply a quantity with a number."""
        if pd.DataFrame and isinstance(other, pd.DataFrame):
            return other.assign(**{col: self * other[col] for col in other.columns})

        if pd.Series and isinstance(other, pd.Series):
            from .pandas_engine import QuantityDtype

            other_unit = getattr(other.dtype, "unit", "")
            val = self * Quantity(other.to_numpy(), other_unit)
            return pd.Series(val.value, index=other.index, dtype=QuantityDtype(unit=val.unit))

        if isinstance(other, float | int | np.ndarray):
            return self.__class__(unit=self.unit, value=self.value * other)

        if isinstance(other, Quantity):
            i_s_units = self.i_s_units * other.i_s_units
            out = self.__class__(
                value=self._to_default_unit() * other._to_default_unit(),
                unit=self.names.get(i_s_units, {}).get("unit", str(i_s_units)),
            )
            output_unit = self._get_output_unit(other, i_s_units)
            if output_unit is not None:
                out = out.to(output_unit)
            return out

        raise NotImplementedError

    def __rmul__(self, other: float | int | float_array | pd.Series | pd.DataFrame) -> Self | pd.Series | pd.DataFrame:
        """Multiply a quantity with a number.
        NOTE: No pandas support for __rmul__.
        """
        if not isinstance(other, float | int | np.ndarray | pd.Series | pd.DataFrame):
            raise TypeError(f"Cannot multiply {other} by quantity.")
        return self.__mul__(other)

    @overload
    def __truediv__(self, other: pd.DataFrame) -> pd.DataFrame: ...
    @overload
    def __truediv__(self, other: pd.Series) -> pd.Series: ...
    @overload
    def __truediv__(self, other: Self | float | int | float_array) -> Self: ...
    def __truediv__(
        self, other: Self | float | int | float_array | pd.Series | pd.DataFrame
    ) -> Self | pd.Series | pd.DataFrame:
        """Divide a quantity with a number or another quantity."""
        if pd.DataFrame and isinstance(other, pd.DataFrame):
            return other.assign(**{col: self / other[col] for col in other.columns})

        if pd.Series and isinstance(other, pd.Series):
            from .pandas_engine import QuantityDtype

            other_unit = getattr(other.dtype, "unit", "")
            val = self / Quantity(other.to_numpy(), other_unit)
            return pd.Series(val.value, index=other.index, dtype=QuantityDtype(unit=val.unit))

        if isinstance(other, float | int | np.ndarray):
            return self.__class__(unit=self.unit, value=self.value / other)

        if isinstance(other, Quantity):
            i_s_units = self.i_s_units / other.i_s_units
            out = self.__class__(
                value=self._to_default_unit() / other._to_default_unit(),
                unit=self.names.get(i_s_units, {}).get("unit", str(i_s_units)),
            )
            output_unit = self._get_output_unit(other, i_s_units)
            if output_unit is not None:
                out = out.to(output_unit)
            return out

        raise NotImplementedError

    def __rtruediv__(self, other: float | int | float_array) -> Self:
        """Divide a number by a quantity.
        NOTE: No pandas support for __rtruediv__.
        """
        if not isinstance(other, float | int | np.ndarray | pd.Series | pd.DataFrame):
            raise TypeError(f"Cannot divide {other} by quantity.")

        i_s_units = self.i_s_units**-1
        return self.__class__(
            value=other / self._to_default_unit(),
            unit=self.names.get(i_s_units, {}).get("unit", str(i_s_units)),
        )

    def __pow__(self, other: int) -> Self:
        """Raise a quantity to a power."""
        if not isinstance(other, int):
            raise TypeError("Can only raise a quantity to an integer power")

        i_s_units = self.i_s_units**other
        return self.__class__(
            value=self._to_default_unit() ** other,
            unit=self.names.get(i_s_units, {}).get("unit", str(i_s_units)),
        )

    def __eq__(self, other) -> bool:
        """Equality."""
        if not isinstance(other, Quantity) or self.i_s_units != other.i_s_units:
            return False
        self_default = self._to_default_unit()
        other_default = other._to_default_unit()
        if isinstance(self_default, float | int):
            return self_default == other_default
        return bool((self_default == other_default).all())

    def __neg__(self) -> Self:
        """Negate."""
        return self.__class__(unit=self.unit, value=-self.value)

    def __repr__(self) -> str:
        """Representation."""
        return f"{self.category}({self})"

    def __str__(self) -> str:
        """String representation."""
        if not isinstance(self.value, float):
            return f"{self.value}, {self.unit}"
        base = f"{self.value:.3f}" if abs(self.value) >= 1e-2 or self.value == 0 else f"{self.value:.2e}"  # noqa: PLR2004
        if "." in base:
            base = base.rstrip("0").rstrip(".")
        return f"{base} {self.unit}" if self.unit else base

    def __len__(self) -> int | None:
        """Length."""
        return None if isinstance(self.value, float | int) else len(self.value)

    def __bool__(self) -> bool:
        """Boolean."""
        return True

    def __iter__(self):
        """Iterator."""
        if isinstance(self.value, float | int):
            yield self
        else:
            for value in self.value:
                yield self.__class__(unit=self.unit, value=value)

    @property
    def dtype(self):
        """Pandas dtype associated with this value."""
        from .pandas_engine import QuantityDtype

        return QuantityDtype(unit=self.unit)

    @classmethod
    def register_unit_rates(
        cls,
        i_s_units: ISUnits | str,
        unit_name: str,
        rate: float = 1.0,
        category: str | None = None,
    ) -> None:
        """Set currency rates."""
        if isinstance(i_s_units, str):
            i_s_units = ISUnits.from_str(i_s_units)

        if i_s_units not in cls.conversion:
            cls.conversion[i_s_units] = {}
        cls.conversion[i_s_units][unit_name] = rate
        if i_s_units not in cls.names:
            if category is None:
                raise ValueError("Category must be specified for new units")
            cls.names[i_s_units] = {"category": category, "unit": unit_name}

        cls.get_unit_info.cache_clear()



from datetime import date, time
from enum import Enum
from numbers import Number
from types import UnionType
from typing import Annotated, Literal, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_utils.common import get_non_null_annotation, is_subclass

SEP = ":"


class Type(Enum):
    """Types of fields."""

    SCALAR = "scalar"
    LITERAL = "literal"
    MODEL = "model"
    UNKNOWN = "unknown"
    DISCRIMINATED_MODEL = "discriminated_model"
    MODEL_LIST = "model_list"
    SCALAR_LIST = "scalar_list"
    LITERAL_LIST = "literal_list"
    UNKOWN_LIST = "unknown_list"
    DISCRIMINATED_MODEL_LIST = "discriminated_model_list"
    MODEL_DICT = "model_dict"
    SCALAR_DICT = "scalar_dict"
    LITERAL_DICT = "literal_dict"
    UNKOWN_DICT = "unkown_dict"
    DISCRIMINATED_MODEL_DICT = "discriminated_model_dict"

    @classmethod
    def classify(cls, annotation: type, discriminator: str | None = None, depth: int = 0) -> bool:  # noqa: PLR0911, PLR0912
        """Classify a value as a field type."""
        annotation = get_non_null_annotation(annotation)

        if get_origin(annotation) is Annotated and get_origin(get_args(annotation)[0]) in [Union, UnionType]:
            if discriminator is None:
                discriminator = next(
                    (f.discriminator for f in get_args(annotation)[1:] if isinstance(f, FieldInfo)), None
                )
            annotation = get_args(annotation)[0]

        if is_subclass(annotation, str | Number | bool | date | time):
            return cls.SCALAR

        if (get_origin(annotation) == Literal) | is_subclass(annotation, Enum):
            return cls.LITERAL

        if is_subclass(annotation, BaseModel):
            return cls.MODEL

        if get_origin(annotation) in [Union, UnionType]:
            if discriminator and all(is_subclass(x, BaseModel) for x in get_args(annotation)):
                return cls.DISCRIMINATED_MODEL
            if all(is_subclass(x, str | Number) for x in get_args(annotation)):
                return cls.SCALAR

        if get_origin(annotation) is list and not depth:
            ann_args = get_args(annotation)
            if not ann_args:
                return cls.UNKOWN_LIST
            args_type = Type.classify(ann_args[0], depth=1)
            if args_type == Type.SCALAR:
                return cls.SCALAR_LIST
            if args_type == Type.LITERAL:
                return cls.LITERAL_LIST
            if args_type == Type.MODEL:
                return cls.MODEL_LIST
            if args_type == Type.DISCRIMINATED_MODEL:
                return cls.DISCRIMINATED_MODEL_LIST
            return cls.UNKOWN_LIST

        if get_origin(annotation) is dict and not depth:
            ann_args = get_args(annotation)
            if not ann_args:
                return cls.UNKOWN_DICT
            args_type = Type.classify(ann_args[1], depth=1)
            if args_type == Type.SCALAR:
                return cls.SCALAR_DICT
            if args_type == Type.LITERAL:
                return cls.LITERAL_DICT
            if args_type == Type.MODEL:
                return cls.MODEL_DICT
            if args_type == Type.DISCRIMINATED_MODEL:
                return cls.DISCRIMINATED_MODEL_DICT
            return cls.UNKOWN_DICT

        return cls.UNKNOWN
